
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

/*
 * Virt Viewer: A virtual machine console viewer
 *
 * Copyright (C) 2007-2012 Red Hat, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Marc-André Lureau <marcandre.lureau@redhat.com>
 */

#ifndef VIRT_VIEWER_ENUMS_H
#define VIRT_VIEWER_ENUMS_H

#include <glib-object.h>
GType virt_viewer_cursor_get_type (void) G_GNUC_CONST;
#define VIRT_VIEWER_TYPE_CURSOR (virt_viewer_cursor_get_type ())
GType virt_viewer_display_show_hint_flags_get_type (void) G_GNUC_CONST;
#define VIRT_VIEWER_TYPE_DISPLAY_SHOW_HINT_FLAGS (virt_viewer_display_show_hint_flags_get_type ())

#endif

/* Generated data ends here */

